import { parser } from 'lezer-dot';
import { syntaxTree, LRLanguage, indentNodeProp, LanguageSupport } from '@codemirror/language';
import { styleTags, tags } from '@lezer/highlight';

// curl 'https://graphviz.org/doc/info/attrs.html'|htmlq 'main table:first-of-type'|htmlq 'td:nth-of-type(1),td:nth-of-type(2),td:nth-of-type(3)' --text|xargs -n3|sed -e's/^/["/g' -e's/$/"],/g' -e's/ /", "/g'|wl-copy
var a = [
    ["_background", "G", "string"],
    ["area", "NC", "double"],
    ["arrowhead", "E", "arrowType"],
    ["arrowsize", "E", "double"],
    ["arrowtail", "E", "arrowType"],
    ["bb", "G", "rect"],
    ["bgcolor", "GC", "colorcolorList"],
    ["center", "G", "bool"],
    ["charset", "G", "string"],
    ["class", "ENCG", "string"],
    ["clusterrank", "G", "clusterMode"],
    ["color", "ENC", "colorcolorList"],
    ["colorscheme", "ENCG", "string"],
    ["comment", "ENG", "string"],
    ["compound", "G", "bool"],
    ["concentrate", "G", "bool"],
    ["constraint", "E", "bool"],
    ["Damping", "G", "double"],
    ["decorate", "E", "bool"],
    ["defaultdist", "G", "double"],
    ["dim", "G", "int"],
    ["dimen", "G", "int"],
    ["dir", "E", "dirType"],
    ["diredgeconstraints", "G", "stringbool"],
    ["distortion", "N", "double"],
    ["dpi", "G", "double"],
    ["edgehref", "E", "escString"],
    ["edgetarget", "E", "escString"],
    ["edgetooltip", "E", "escString"],
    ["edgeURL", "E", "escString"],
    ["epsilon", "G", "double"],
    ["esep", "G", "addDoubleaddPoint"],
    ["fillcolor", "NEC", "colorcolorList"],
    ["fixedsize", "N", "boolstring"],
    ["fontcolor", "ENGC", "color"],
    ["fontname", "ENGC", "string"],
    ["fontnames", "G", "string"],
    ["fontpath", "G", "string"],
    ["fontsize", "ENGC", "double"],
    ["forcelabels", "G", "bool"],
    ["gradientangle", "NCG", "int"],
    ["group", "N", "string"],
    ["head_lp", "E", "point"],
    ["headclip", "E", "bool"],
    ["headhref", "E", "escString"],
    ["headlabel", "E", "lblString"],
    ["headport", "E", "portPos"],
    ["headtarget", "E", "escString"],
    ["headtooltip", "E", "escString"],
    ["headURL", "E", "escString"],
    ["height", "N", "double"],
    ["href", "GCNE", "escString"],
    ["id", "GCNE", "escString"],
    ["image", "N", "string"],
    ["imagepath", "G", "string"],
    ["imagepos", "N", "string"],
    ["imagescale", "N", "boolstring"],
    ["inputscale", "G", "double"],
    ["K", "GC", "double"],
    ["label", "ENGC", "lblString"],
    ["label_scheme", "G", "int"],
    ["labelangle", "E", "double"],
    ["labeldistance", "E", "double"],
    ["labelfloat", "E", "bool"],
    ["labelfontcolor", "E", "color"],
    ["labelfontname", "E", "string"],
    ["labelfontsize", "E", "double"],
    ["labelhref", "E", "escString"],
    ["labeljust", "GC", "string"],
    ["labelloc", "NGC", "string"],
    ["labeltarget", "E", "escString"],
    ["labeltooltip", "E", "escString"],
    ["labelURL", "E", "escString"],
    ["landscape", "G", "bool"],
    ["layer", "ENC", "layerRange"],
    ["layerlistsep", "G", "string"],
    ["layers", "G", "layerList"],
    ["layerselect", "G", "layerRange"],
    ["layersep", "G", "string"],
    ["layout", "G", "string"],
    ["len", "E", "double"],
    ["levels", "G", "int"],
    ["levelsgap", "G", "double"],
    ["lhead", "E", "string"],
    ["lheight", "GC", "double"],
    ["lp", "EGC", "point"],
    ["ltail", "E", "string"],
    ["lwidth", "GC", "double"],
    ["margin", "NCG", "doublepoint"],
    ["maxiter", "G", "int"],
    ["mclimit", "G", "double"],
    ["mindist", "G", "double"],
    ["minlen", "E", "int"],
    ["mode", "G", "string"],
    ["model", "G", "string"],
    ["mosek", "G", "bool"],
    ["newrank", "G", "bool"],
    ["nodesep", "G", "double"],
    ["nojustify", "GCNE", "bool"],
    ["normalize", "G", "doublebool"],
    ["notranslate", "G", "bool"],
    ["nslimit", "G", "double"],
    ["nslimit1", "G", "double"],
    ["ordering", "GN", "string"],
    ["orientation", "NG", "doublestring"],
    ["outputorder", "G", "outputMode"],
    ["overlap", "G", "stringbool"],
    ["overlap_scaling", "G", "double"],
    ["overlap_shrink", "G", "bool"],
    ["pack", "G", "boolint"],
    ["packmode", "G", "packMode"],
    ["pad", "G", "doublepoint"],
    ["page", "G", "doublepoint"],
    ["pagedir", "G", "pagedir"],
    ["pencolor", "C", "color"],
    ["penwidth", "CNE", "double"],
    ["peripheries", "NC", "int"],
    ["pin", "N", "bool"],
    ["pos", "EN", "pointsplineType"],
    ["quadtree", "G", "quadTypebool"],
    ["quantum", "G", "double"],
    ["rank", "S", "rankType"],
    ["rankdir", "G", "rankdir"],
    ["ranksep", "G", "doubledoubleList"],
    ["ratio", "G", "doublestring"],
    ["rects", "N", "rect"],
    ["regular", "N", "bool"],
    ["remincross", "G", "bool"],
    ["repulsiveforce", "G", "double"],
    ["resolution", "G", "double"],
    ["root", "GN", "stringbool"],
    ["rotate", "G", "int"],
    ["rotation", "G", "double"],
    ["samehead", "E", "string"],
    ["sametail", "E", "string"],
    ["samplepoints", "N", "int"],
    ["scale", "G", "doublepoint"],
    ["searchsize", "G", "int"],
    ["sep", "G", "addDoubleaddPoint"],
    ["shape", "N", "shape"],
    ["shapefile", "N", "string"],
    ["showboxes", "ENG", "int"],
    ["sides", "N", "int"],
    ["size", "G", "doublepoint"],
    ["skew", "N", "double"],
    ["smoothing", "G", "smoothType"],
    ["sortv", "GCN", "int"],
    ["splines", "G", "boolstring"],
    ["start", "G", "startType"],
    ["style", "ENCG", "style"],
    ["stylesheet", "G", "string"],
    ["tail_lp", "E", "point"],
    ["tailclip", "E", "bool"],
    ["tailhref", "E", "escString"],
    ["taillabel", "E", "lblString"],
    ["tailport", "E", "portPos"],
    ["tailtarget", "E", "escString"],
    ["tailtooltip", "E", "escString"],
    ["tailURL", "E", "escString"],
    ["target", "ENGC", "escStringstring"],
    ["tooltip", "NECG", "escString"],
    ["truecolor", "G", "bool"],
    ["URL", "ENGC", "escString"],
    ["vertices", "N", "pointList"],
    ["viewport", "G", "viewPort"],
    ["voro_margin", "G", "double"],
    ["weight", "E", "intdouble"],
    ["width", "N", "double"],
    ["xdotversion", "G", "string"],
    ["xlabel", "EN", "lblString"],
    ["xlp", "NE", "point"],
    ["z", "N", "double"],
];

// curl 'https://graphviz.org/docs/attr-types/arrowType/' | htmlq 'main table code' --text | sed -e's/^/"/g' -e's/$/",/g'|wl-copy
var arrowtype = [
    "normal",
    "inv",
    "dot",
    "invdot",
    "odot",
    "invodot",
    "none",
    "tee",
    "empty",
    "invempty",
    "diamond",
    "odiamond",
    "ediamond",
    "crow",
    "box",
    "obox",
    "open",
    "halfopen",
    "vee",
];

var pagedir = [
    '"BL"',
    '"BR"',
    '"TL"',
    '"TR"',
    '"RB"',
    '"RT"',
    '"LB"',
    '"LT"',
];

// curl 'https://graphviz.org/doc/info/shapes.html'|htmlq 'figcaption code' --text|sed -e's/^/"/g' -e's/$/",/g'|wl-copy
const polygonbased = [
    "box",
    "polygon",
    "ellipse",
    "oval",
    "circle",
    "point",
    "egg",
    "triangle",
    "plaintext",
    "plain",
    "diamond",
    "trapezium",
    "parallelogram",
    "house",
    "pentagon",
    "hexagon",
    "septagon",
    "octagon",
    "doublecircle",
    "doubleoctagon",
    "tripleoctagon",
    "invtriangle",
    "invtrapezium",
    "invhouse",
    "Mdiamond",
    "Msquare",
    "Mcircle",
    "rect",
    "rectangle",
    "square",
    "star",
    "none",
    "underline",
    "cylinder",
    "note",
    "tab",
    "folder",
    "box3d",
    "component",
    "promoter",
    "cds",
    "terminator",
    "utr",
    "primersite",
    "restrictionsite",
    "fivepoverhang",
    "threepoverhang",
    "noverhang",
    "assembly",
    "signature",
    "insulator",
    "ribosite",
    "rnastab",
    "proteasesite",
    "proteinstab",
    "rpromoter",
    "rarrow",
    "larrow",
    "lpromoter",
];
const recordbased = [
    "record",
    "Mrecord",
];
var shape = /*@__PURE__*/polygonbased.concat(recordbased);

var smoothtype = [
    '"none"',
    '"avg_dist"',
    '"graph_dist"',
    '"power_dist"',
    '"rng",',
    '"spring"',
    '"triangle"',
];

const nodesedges = [
    '"dashed"',
    '"dotted"',
    '"solid"',
    '"invis"',
    '"bold"',
];
const edges = [
    '"tapered"',
];
const nodes = [
    '"filled"',
    '"striped"',
    '"wedged"',
    '"diagonals"',
    '"rounded"',
];
const clusters = [
    '"filled"',
    '"striped"',
    '"rounded"',
];
var style = /*@__PURE__*/nodesedges.concat(edges).concat(nodes).concat(clusters);

class Attr {
    constructor(name, usedby, type) {
        this.name = name;
        this.usedby = new Set(Array.from(usedby, v => {
            switch (v) {
                case "E":
                case "N":
                case "G":
                case "S":
                case "C":
                    return v;
                default:
                    throw new Error(`invalid value ${v}`);
            }
        }));
        this.type = type;
    }
    intoCompletion() {
        return {
            label: this.name,
            type: "property",
            detail: this.type,
        };
    }
}
const attrs = /*@__PURE__*/a.map(([name, usedby, type]) => new Attr(name, usedby, type));
const gattrs = /*@__PURE__*/attrs.filter(a => a.usedby.has("G")).map(a => a.intoCompletion());
const eattrs = /*@__PURE__*/attrs.filter(a => a.usedby.has("E")).map(a => a.intoCompletion());
const nattrs = /*@__PURE__*/attrs.filter(a => a.usedby.has("N")).map(a => a.intoCompletion());
const attrbyname = /*@__PURE__*/new Map(/*@__PURE__*/attrs.map(a => [a.name, a]));
function pos(cx, near) {
    if (cx.pos !== near.to) {
        return cx.pos;
    }
    switch (near.name) {
        case "Simpleid":
        case "Quoted":
            return near.from;
        default:
            return cx.pos;
    }
}
function result(cx, near, options) {
    return {
        from: pos(cx, near),
        options,
        //span: /\w*$/, // TODO quoted token
    };
}
function completeAttrName(cx, near, node) {
    var _a;
    let current = node;
    while ((current = current.parent) !== null) {
        switch (current.name) {
            case "AttrStmt":
                switch ((_a = current.firstChild) === null || _a === void 0 ? void 0 : _a.name) {
                    case "edge":
                        return result(cx, near, eattrs);
                    case "graph":
                        return result(cx, near, gattrs);
                    case "node":
                        return result(cx, near, nattrs);
                    default:
                        return null;
                }
            case "NodeStmt":
                return result(cx, near, nattrs);
            case "EdgeStmt":
                return result(cx, near, eattrs);
        }
    }
    return null;
}
function completeByType(cx, near, type) {
    function r(labels) {
        return result(cx, near, labels.map(label => Object.assign({ label, type: "text" })));
    }
    switch (type) {
        case "arrowType":
            return r(arrowtype);
        case "bool":
            return r(["true", "false"]);
        case "clusterMode":
            return r(['"local"', '"global"', '"none"']);
        case "dirType":
            return r(['"forward"', '"back"', '"both"', '"none"']);
        case "portPos":
            return r(["n", "ne", "e", "se", "s", "sw", "w", "nw", "c", "_"]);
        case "outputMode":
            return r(['"breadthfirst"', '"nodesfirst"', '"edgesfirst"']);
        case "packMode":
            return r(['"node"', '"clust"', '"graph"', '"array"']); // FIXME array variant
        case "pagedir":
            return r(pagedir);
        case "rankType":
            return r(['"same"', '"min"', '"source"', '"max"', '"sink"']);
        case "rankdir":
            return r(['"TR"', '"LR"', '"BT"', '"RL"']);
        case "shape":
            return r(shape);
        case "smoothType":
            return r(smoothtype);
        case "style":
            // TODO contextual
            return r(style);
        default:
            return null;
    }
}
function completeAttrVal(cx, near) {
    const cursor = near.cursor();
    while (true) {
        if (cursor.name === "=") {
            break;
        }
        if (!cursor.prev()) {
            return null;
        }
    }
    while (cursor.prev()) {
        switch (cursor.node.name) {
            case "Simpleid": {
                const { from, to } = cursor.node;
                const name = cx.state.sliceDoc(from, to);
                const attr = attrbyname.get(name);
                if (attr) {
                    return completeByType(cx, near, attr.type);
                }
                break;
            }
            case "Quoted":
            case "Htmlstr":
                // Not supported.
                return null;
        }
    }
    return null;
}
function completeEdgeop(cx, tree, near) {
    const cursor = tree.cursor();
    while (cursor.next()) {
        switch (cursor.node.name) {
            case "graph":
                return result(cx, near, [{ label: "-" }]);
            case "digraph":
                return result(cx, near, [{ label: ">" }]);
        }
    }
    return null;
}
function completeNodeIds(cx, tree) {
    const ids = [];
    for (const cursor = tree.cursor(); cursor.next();) {
        if (cursor.node.name === 'NodeId') {
            const { from, to } = cursor.node;
            if (!(from <= cx.pos && to >= cx.pos)) {
                const id = cx.state.doc.sliceString(from, to);
                ids.push({
                    label: id,
                    type: "variable",
                    boost: 1,
                });
            }
        }
    }
    return ids;
}
function completeInAttrList(cx, near, current) {
    if (near.name === "=") {
        return completeAttrVal(cx, near);
    }
    return completeAttrName(cx, near, current);
}
function completeInNodeId(cx, tree, near) {
    const ids = completeNodeIds(cx, tree);
    return result(cx, near, ids);
}
function completeInEdgeStmt(cx, tree, near) {
    if (near.name !== "edgeop") {
        return null;
    }
    const ids = completeNodeIds(cx, tree);
    return result(cx, near, ids);
}
function completeInProp(cx, near) {
    if (near.name === "=") {
        return completeAttrVal(cx, near);
    }
    return null;
}
function completeInStmtList(cx, tree, near) {
    const ids = completeNodeIds(cx, tree);
    // completion keyword and property name explicit only.
    if (cx.explicit) {
        const options = ids.concat([
            { label: "node", type: "keyword" },
            { label: "graph", type: "keyword" },
            { label: "edge", type: "keyword" },
            { label: "subgraph", type: "keyword" },
        ]).concat(gattrs);
        return result(cx, near, options);
    }
    else {
        return result(cx, near, ids);
    }
}
function completeInGraph(cx, near) {
    const cursor = near.cursor();
    let foundStrict = false;
    while (cursor.prev()) {
        switch (cursor.node.name) {
            case "graph":
            case "digraph":
                return null;
            case "strict":
                foundStrict = true;
                break;
        }
    }
    return result(cx, near, foundStrict ? [
        { label: "digraph", type: "keyword" },
        { label: "graph", type: "keyword" },
    ] : [
        { label: "digraph", type: "keyword" },
        { label: "graph", type: "keyword" },
        { label: "strict", type: "keyword", boost: -1 },
    ]);
}
/*
 * resolve node
 *
 * ```
 * graph { a[label = │] }
 *                   ^-- cursor
 *                 ^---- resolve
 * ```
 */
function nearNode(cx, tree) {
    //const comment = ["LineComment", "SharpComment", "BlockComment"];
    const skipToken = /[ \t\r\n]*/;
    const matches = cx.matchBefore(skipToken);
    const pos = matches !== null ? matches.from : 0; // 0 ... correct ??
    const node = tree.resolveInner(pos, -1);
    // FIXME skip comments
    return node;
}
function complete(debug, cx) {
    var _a;
    const tree = syntaxTree(cx.state);
    const near = nearNode(cx, tree);
    if (debug) {
        (() => {
            const stack = [];
            for (let node = near; node; node = node.parent) {
                stack.push(node.name);
            }
            console.log(stack.join("/"));
        })();
    }
    for (let node = near; node; node = node.parent) {
        switch (node.name) {
            case "graph":
            case "digraph":
            case "strict":
            case "subgraph":
            case "LineComment":
            case "SharpComment":
            case "BlockComment":
                return null;
            case "{":
            case ";":
                if (cx.pos === near.to && !cx.explicit) {
                    // without space. ignore completion.
                    // graph {│}
                    return null;
                }
                break;
            case "AttrVal":
                if (cx.pos === near.to) {
                    // without space.
                    // graph { a[shape = b│] }
                    //                    ^-- cursor
                    //                   ^--- completion target
                    return completeAttrVal(cx, near);
                }
                // skip this case.
                // graph { a[shape = b │] }
                //                     ^-- cursor
                //                    ^--- space
                break;
            case "NodeId":
                if (cx.pos === node.to && ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.name) !== "NodeStmt") {
                    // without space.
                    // graph { aaa; aaa -- a│}
                    //                      ^-- cursor
                    //                     ^--- completion target
                    return completeInNodeId(cx, tree, near);
                }
                // skip this case.
                // graph { aaa; aaa -- a │}
                //                       ^-- cursor
                //                      ^--- space
                //
                // skip because ambiguity.
                // graph { rrr; r│ } .. maybe enter `rankdir = ..`
                break;
            case "Prop":
                // o `graph { rankdir = │ }` .. =/**Prop**/StmtList/Graph
                // x `graph { r│ }` .. Simpleid/ID/**NodeId**/NodeStmt/StmtList/Graph
                // x `graph { rankdir = L│ }` .. Simpleid/ID/**AttrVal**/Prop/StmtList/Graph
                return completeInProp(cx, near);
            case "AttrList":
                // graph { graph[│] } .. [/**AttrList**/AttrStmt/StmtList/Graph
                // graph { graph[r│] } .. Simpleid/ID/AttrName/Attr/**AttrList**/AttrStmt/StmtList/Graph
                // graph { graph[rankdir =│] } .. =/Attr/**AttrList**/AttrStmt/StmtList/Graph TODO
                return completeInAttrList(cx, near, node);
            case "EdgeStmt":
                // x graph { l│ } .. Simpleid/ID/**NodeId**/NodeStmt/StmtList/Graph
                // x graph { l -│ } .. ⚠/NodeId/NodeStmt/**StmtList**/Graph
                // o graph { l --│ } .. edgeop/**EdgeStmt**/StmtList/Graph
                // x graph { l -- r│ } Simpleid/ID/**NodeId**/EdgeStmt/StmtList/Graph
                return completeInEdgeStmt(cx, tree, near);
            case "StmtList":
                if (cx.pos === near.to && cx.state.sliceDoc(near.from, near.to) === "-") {
                    // graph { a - } .. ⚠/NodeId/NodeStmt/**StmtList**/Graph
                    return completeEdgeop(cx, tree, near);
                }
                // graph { │ } .. {/**StmtList**/Graph
                // graph { n│ } .. Simpleid/ID/NodeId/NodeStmt/**StmtList**/Graph
                return completeInStmtList(cx, tree, near);
            case "Graph":
                // │ .. Graph
                return completeInGraph(cx, near);
        }
    }
    return null;
}

const dotLanguage = /*@__PURE__*/LRLanguage.define({
    parser: /*@__PURE__*/parser.configure({
        props: [
            /*@__PURE__*/indentNodeProp.add({
                StmtList: cx => cx.baseIndent + cx.unit,
            }),
            /*@__PURE__*/styleTags({
                "graph digraph node edge strict subgraph": tags.keyword,
                Quoted: tags.string,
                Numeral: tags.number,
                ID: tags.variableName,
                "edgeop =": tags.operator,
                "[ ]": tags.squareBracket,
                "{ }": tags.brace,
                ", ;": tags.separator,
                "LineComment SharpComment": tags.lineComment,
                BlockComment: tags.blockComment,
            }),
        ],
    }),
    languageData: {
        closeBrackets: { brackets: ["[", "{", '"'] },
        commentTokens: { line: "//", block: { open: "/*", close: "*/" } },
        indentOnInput: /^\s*(?:\{|\})$/,
    },
});
function dot(config = {}) {
    var _a;
    return new LanguageSupport(dotLanguage, dotLanguage.data.of({
        autocomplete: complete.bind(null, ((_a = config.debug) !== null && _a !== void 0 ? _a : false)),
    }));
}

export { dot, dotLanguage };
